#!/usr/bin/env python3

"""
EmPy documentation generating system.  This module requires a modern
Python 3.x interpreter.
"""

#
# imports
#

import hashlib
import io
import os
import subprocess
import sys
import time

import em
import emhelp
import emlib

#
# constants
#

SOURCE = '\N{KEYBOARD}\ufe0f'
OUTPUT = '\N{DESKTOP COMPUTER}\ufe0f'

PATHS = ['../..', '.']

EMOJIS = {
    '...': ('\N{HORIZONTAL ELLIPSIS}', "horizontal ellipsis"),
    '<--': ('\N{LONG LEFTWARDS ARROW}', "long leftwards arrow"),
    '-->': ('\N{LONG RIGHTWARDS ARROW}', "long rightwards arrow"),
}

def admonish(text, emoji=None, admonition="important"):
    """Create an admonition."""
    if emoji is None:
        return ":::{{{admonition}}}\n{text}\n:::\n".format(
            admonition=admonition, text=text)
    else:
        return ":::{{{admonition}}}\n{emoji} {text}\n:::\n".format(
            admonition=admonition, emoji=emoji, text=text)

if not os.environ.get('EMPY_STRICT', None):
    # Only make the alerting admonitions available available if this is not
    # strict; if strict, this will result in an emoji error.
    EMOJIS.update({
        '!!!': (admonish("!!!", "\u2757\ufe0f"), "exclamation mark"),
        '???': (admonish("???", "\u2753\ufe0f"), "question mark"),
        '^^^': (admonish("^^^", "\u26a0\ufe0f"), "warning sign"),
        '///': (admonish("///", "\u2714\ufe0f"), "check mark"),
        '\\\\\\': (admonish("\\\\\\", "\u274c\ufe0f"), "cross mark"),
        '+++': (admonish("+++", "\U0001f53a"), "red triangle pointed up"),
        '---': (admonish("---", "\U0001f53b"), "red triangle pointed down"),
    })

#
# Error ...
#

class DocumentationError(em.Error): pass

#
# Identity
#

class Identity:

    """Dynamically access magic attributes on both the interpreter and
    underlying module."""

    def __init__(self, interp, module):
        self.interp = interp
        self.module = module

    def __str__(self):
        try:
            return self.__getattr__('project')
        except AttributeError:
            return self.__getattr__('program')

    def __getattr__(self, attr):
        attribute = '__{}__'.format(attr)
        if attribute in self.interp.globals:
            return self.interp.globals[attribute]
        elif attribute in self.module.__dict__:
            return self.module.__dict__[attribute]
        else:
            raise AttributeError("unknown attribute: {}".format(__attribute__))

    def tarball(self, version='latest'):
        if version is None:
            version = self.version
        return '{}{}-{}.tar.gz'.format(self.url, self.program, version)

    def path(self, suffix=None):
        if suffix is not None:
            return self.url + suffix
        else:
            return self.url

#
# Tee
#

class Tee:

    """A file-like object which can split output into multiple files."""

    temporary = '/tmp'

    def __init__(self, files):
        self.files = files

    def __call__(self, string):
        self.write(string)

    def write(self, string):
        for file in self.files:
            file.write(string)

    def writelines(self, lines):
        for line in lines:
            self.write(line)

#
# Information
#

class Information:

    """A helper information class to generate documentation."""

    timeFormat = '%Y-%m-%d %H:%M:%S'
    extensions = ['.md.em', '.md.pre', '.md', '']
    hashFactory = hashlib.sha1
    hashName = 'SHA1'
    encoding = 'utf-8'

    class Flag:

        """Encapsulate a flag:  a command line option, a configuration
        variable, and/or an environment variable."""

        def __init__(self, config, options,
                     var=None, env=None, val=None):
            self.config = config
            self.options = options
            self.var = var
            self.val = val
            self.env = env

        def __str__(self):
            return str((self.options, self.var, self.val, self.env))

        def render(self, file, key,
                   showDefault=True, default=None, verbose=False):
            isOption = key.startswith('-')
            options = '/'.join("{}".format(x) for x in self.options)
            environ = '{}'.format(self.env) if self.env else None
            if self.var is not None:
                if (verbose and not showDefault and
                    self.val is not None):
                    variable = '{}\N{NO-BREAK SPACE}=\N{NO-BREAK SPACE}{}'.format(
                        self.var, repr(self.val))
                else:
                    variable = '{}'.format(self.var)
            else:
                variable = None
            declarations = []
            fragments = []
            if isOption:
                assert options, (key, str(self))
                # It's an option, so list those first.
                declarations.append(options)
                if environ:
                   fragments.extend(["_environment variable:_ `{}`".format(
                       environ)])
                if variable:
                    fragments.extend(["_configuration variable:_ `{}`".format(
                        variable)])
            elif key.startswith('EMPY_'):
                assert environ, (key, str(self))
                # If it starts with EMPY_, it's an environment variable.
                declarations.append(environ)
                if options:
                    fragments.extend(["_command line option:_ `{}`".format(
                        options)])
                if variable:
                    fragments.extend(["_configuration variable:_ `{}`".format(
                        variable)])
            else:
                assert variable, (key, str(self))
                # Otherwise, it's a configuration variable.
                declarations.append(variable)
                types = self.config._specs[key]
                if types is not None and verbose:
                    if types == em.strType:
                        types = str
                    if types is list:
                        types = 'list[str]'
                    if not isinstance(types, tuple):
                        types = (types,)
                    optional = False
                    if types[-1] is None:
                        # If the last one is None, then the type is
                        # Optional.
                        optional = True
                        types = types[:-1]
                    types = ' | '.join(
                        x if isinstance(x, str) else x.__name__
                        for x in types)
                    if optional:
                        types = 'Optional[%s]' % types
                    declarations.append(": {}".format(types))
                if options:
                    fragments.extend(["_command line option:_ `{}`".format(options)])
                if environ:
                   fragments.extend(["_environment variable:_ `{}`".format(environ)])
            if showDefault:
                declarations.append("\N{NO-BREAK SPACE}=\N{NO-BREAK SPACE}{}".format(default))
            if fragments and verbose:
                file.write('`{}` ({})'.format(''.join(declarations),
                                              ', '.join(fragments)))
            else:
                file.write('`{}`'.format(''.join(declarations)))

    noLanguage = 'text'

    def __init__(self, interp, moduleName, file=sys.stdout):
        self.interp = interp
        self.moduleName = moduleName
        self.file = file
        self.module = __import__(moduleName)
        self.ident = Identity(interp, self.module)
        self.details = emlib.Details()
        self.config = em.Configuration()
        self.usage = emhelp.Usage(config=self.config)
        self.options = self.process()

    def process(self):
        options = {}
        section = self.usage.payload['options']
        for entry in section.entries:
            fullOptions = [x for x in entry.raw if not x.isspace()]
            theseOptions = [x.split('=', 1)[0] for x in fullOptions]
            flag = self.Flag(self.config, fullOptions,
                var=entry.var, val=entry.val, env=entry.env)
            for key in theseOptions:
                assert key not in options, key
                options[key] = flag
            if entry.var:
                options[entry.var] = flag
            if entry.env:
                options[entry.env.split(' ', 1)[0]] = flag
        for key, value in self.config.__dict__.items():
            if key in options:
                options[key].default = value
            else:
                flag = self.Flag(self.config, [], var=key)
                options[key] = flag
        return options

    def topic(self, topic, separator=False):
        self.file.write("```{}\n".format(self.noLanguage))
        self.usage.show([topic], separator)
        self.file.write("```\n")

    def option(self, key, verbose=False, showDefault=False):
        if verbose is None:
            self.file.write('`{}`'.format(key))
            return
        if showDefault and key in self.config:
            if key == 'pathSeparator':
                default = "';'` (Windows) or `':'` (others) ` "
            else:
                default = getattr(self.config, key)
                if isinstance(default, dict):
                    default = "{...}"
                else:
                    default = repr(default)
        else:
            default = None
        self.options[key].render(self.file, key, showDefault, default, verbose)

    def variable(self, variable):
        self.file.write("`{}`".format(variable))

    def source(self, filename):
        for extension in self.extensions:
            if os.path.exists(filename + extension):
                return filename + extension
        else:
            raise FileNotFoundError("source file not found: {}".format(filename))

    def filter(self, text, lines=None, blanks=None):
        if isinstance(text, bytes) and self.encoding:
            text = text.decode(self.encoding)
        text = (text
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;'))
        stopped = False
        chunks = []
        for line in text.splitlines():
            if line.startswith('...'):
                chunks.append("<i>{}</i>".format(line))
            else:
                chunks.append(line)
            if not line and blanks is not None:
                blanks -= 1
                if blanks == 0:
                    stopped = True
            if lines is not None and len(chunks) >= lines:
                stopped = True
            if stopped:
                break
        if stopped:
            chunks.append('...')
        return '\n'.join(chunks)

    def shell(self, command, output, prefix='% ', lines=None, blanks=None,
              class_='shell', exitCode=0):
        display = command
        if isinstance(display, list):
            words = ['"{}"'.format(x)
                if ' ' in x and not x.startswith('#')
                else x
                for x in command]
            display = ' '.join(words)
        self.file.write("<pre class=\"{}\">\n".format(class_))
        if display:
            self.file.write("<b><i>{}{}</i></b>\n".format(
                prefix, self.filter(display)))
        self.file.write(self.filter(output, lines, blanks))
        if exitCode != 0:
            self.file.write("<i>Exit code: {}</i>\n".format(exitCode))
        self.file.write("</pre>\n")

    def execute(self, command, prefix='% ', lines=None, blanks=None, check=True):
        oldPath = os.environ['PATH']
        try:
            # Make sure that if it's em.py, it's the local one.
            path = '.:' + oldPath
            os.environ['PATH'] = path
            result = subprocess.run(command, capture_output=True)
            output = result.stdout
            self.shell(command, output, prefix, lines, blanks,
                       exitCode=result.returncode if check else 0)
        finally:
            os.environ['PATH'] = oldPath

    def splice(self, file, name='<splice>'):
        context = self.interp.newContext(name)
        self.interp.pushContext(context)
        try:
            self.interp.fileFull(file)
        finally:
            self.interp.popContext()

    def load(self, filename, mode='r'):
        with open(filename, mode) as file:
            self.splice(file, filename)

    def clip(self, filename, heading, rename=None, stoplines=None, mode='r'):
        buffer = []
        on = False
        start = 0
        number = 1
        if stoplines is None:
            stoplines = []
        stoplines = [x if x.endswith('\n') else x + '\n' for x in stoplines]
        with open(filename, mode) as file:
            for line in file.readlines():
                if line.startswith('#'):
                    try:
                        prelim, title = line.split(' ', 1)
                    except ValueError:
                        raise DocumentationError("malformed header line: {}".format(line), line=line)
                    if ':' in title:
                        title, subtitle = title.split(':', 1)
                        subtitle = subtitle.strip()
                    title = title.strip()
                    level = prelim.count('#')
                    if title == heading:
                        if rename is not None:
                            line = ('#' * level) + ' ' + rename + '\n'
                        if not on:
                            start = number
                            on = True
                    else:
                        if on:
                            break
                if line in stoplines:
                    on = False
                if on:
                    buffer.append(line)
                number += 1
        if start:
            self.splice(io.StringIO(''.join(buffer)), (filename, start))
        else:
            raise DocumentationError("could not find heading '{}'".format(heading), heading=heading)

    def tee(self, filename):
        return Tee(filename)

    def summarize(self, filename='README'):
        filename = self.source(filename)
        hasher = self.hashFactory()
        with open(filename, 'rb') as f:
            data = f.read()
        length = len(data)
        hasher.update(data)
        hash = hasher.hexdigest()
        record = os.stat(filename)
        when = time.localtime(record.st_mtime)
        timeStamp = time.strftime(self.timeFormat, when)
        out = em.StringIO('w')
        self.details.show(file=out)
        self.file.write("""\
_This documentation for {} version {} was generated from {} ({} `{}`, {} bytes) at {} using {}._
""".format(
    self.ident, self.ident.version, filename,
    self.hashName, hash, length,
    timeStamp, out.getvalue()))
        self.done()

    def done(self):
        if self.config:
            self.interp.dropAllDiversions()
            self.config = None

#
# Extension
#

class Extension(em.Extension):

    """The EmPy documentation extension."""

    languages = {
        'empy': '' # to eliminate Pygments warning
    }

    sub = True

    def __init__(self):
        super(Extension, self).__init__()
        self.current = 0

    def next(self, amount=1):
        self.current += amount
        return self.current

    def expand(self, source, root='<root>'):
        with em.StringIO() as file:
            with em.Interpreter(output=file, root=root) as interp:
                interp.string(source)
            output = file.getvalue()
        return output

    def angle_brackets(self, source, depth, locals):
        caption = None
        if source.startswith('['):
            language = 'empy'
            caption, source = source[1:].split(']', 1)
            example = True
        else:
            if not source.startswith('\n'):
                language, source = source.split('\n', 1)
                language = language.strip()
            else:
                language = ''
            example = False
        if caption:
            suffix = ': ' + caption
        else:
            suffix = ''
        source = source.strip() + '\n'
        if example:
            number = self.next()
            self.interp.startDiversion(caption)
            self.interp.write('<a class="example" id="example-{}"></a>\n'.format(number))
            self.interp.write('\n')
            self.interp.write(':::{{admonition}} Example {}{}\n'.format(
                number, suffix))
            output = self.expand(source, '<example {}{}>'.format(
                number, ' "' +  caption + '"' if caption else ''))
            self.interp.write('_Source_: {}\n'.format(SOURCE))
            self.interp.write('<div class="source">\n\n')
            self.interp.write('``````{}\n'.format(self.languages.get(language, language)))
            self.interp.write(source)
            self.interp.write('``````\n')
            self.interp.write('\n</div>\n\n')
            self.interp.write('_Output_: {}\n'.format(OUTPUT))
            self.interp.write('<div class="output">\n\n')
            self.interp.write('``````\n')
            self.interp.write(output)
            self.interp.write('``````\n')
            self.interp.write('\n</div>\n\n')
            self.interp.write(':::\n')

            self.interp.stopDiverting()
            self.interp.replayDiversion(caption)
        else:
            self.interp.write('``````{}\n'.format(self.languages.get(language, language)))
            self.interp.write(source)
            self.interp.write('``````\n')

#
# Hook
#

class Hook(emlib.Hook):

    """The EmPy documentation hook."""

    delimiter = '`'

    def __init__(self, interp):
        self.interp = interp

    def preBackquote(self, literal):
        if self.delimiter in literal:
            for count in range(1, 10):
                if '`' * count not in literal:
                    count -= 1
                    break
        else:
            count = 0
        affix = self.delimiter * (count + 1)
        self.interp.write("{}{}{}".format(affix, literal, affix))
        return True

#
# init
#

def init(interp, moduleName, paths=None):
    """Initialize the information object."""
    if paths is None:
        paths = PATHS
    for path in paths:
        sys.path.insert(0, os.path.abspath(path))
    interp.config.emojis = EMOJIS
    interp.addHook(Hook(interp))
    interp.installExtension(Extension())
    return Information(interp, moduleName)



#!/usr/bin/env python3

"""
Optional support classes for EmPy.
"""

#
# imports
#

import operator
import os
import platform
import sys

import em

#
# Filter ...
#

class Filter(em.Root):

    """An abstract filter."""

    # Meta methods.

    def follow(self):
        """Return the next filter/file-like object in the sequence, or
        None."""
        raise NotImplementedError

    def isAttached(self):
        """Is a filter/file already attached to this one?"""
        raise NotImplementedError

    def attach(self, filter):
        """Attach a filter to this one."""
        raise NotImplementedError

    def detach(self):
        """Detach a filter from its sink."""
        raise NotImplementedError

    def last(self):
        """Find the last filter in this chain."""
        this, last = self, None
        while this is not None:
            last = this
            this = this.follow()
        return last

    def _flush(self):
        """The _flush method should always flush the underlying sinks."""
        raise NotImplementedError

    # File-like methods.

    def write(self, data):
        """The write method must be implemented."""
        raise NotImplementedError

    def writelines(self, lines):
        """Write lines.  This should not need to be overriden
        for most filters."""
        for line in lines:
            self.write(line)

    def flush(self):
        """The flush method can be overridden."""
        self._flush()

    def close(self):
        """The close method must be implemented."""
        raise NotImplementedError


class NullFilter(Filter):

    """A filter that never sends any output to its sink."""

    def follow(self): return None
    def isAttached(self): return False
    def attach(self, filter): pass
    def detach(self): pass

    def write(self, data): pass
    def close(self): pass


class ConcreteFilter(Filter):

    """A concrete filter has a single sink."""

    def __init__(self):
        if self.__class__ is ConcreteFilter:
            raise NotImplementedError
        self.sink = None

    def follow(self):
        """Return the next filter/file-like object in the sequence, or None."""
        return self.sink

    def isAttached(self):
        """Is a filter/file already attached to this one?"""
        return self.sink is not None

    def attach(self, filter):
        """Attach a filter to this one."""
        if self.sink is not None:
            # If one's already attached, detach it first.
            self.detach()
        self.sink = filter

    def detach(self):
        """Detach a filter from its sink."""
        self.flush()
        self._flush() # do a guaranteed flush to just to be safe
        self.sink = None

    def last(self):
        """Find the last filter in this chain."""
        this, last = self, None
        while this is not None:
            last = this
            this = this.follow()
        return last

    def _flush(self):
        """This method should flush the concrete sink and should
        not be overridden."""
        self.sink.flush()

    def flush(self):
        self._flush()

    def close(self):
        """Close the filter.  Do an explicit flush first, then close the
        sink."""
        self.flush()
        self.sink.close()


class IdentityFilter(ConcreteFilter):

    """A filter which just sends any output straight through to its sink."""

    def write(self, data):
        self.sink.write(data)


class FunctionFilter(ConcreteFilter):

    """A filter that works simply by pumping its input through a
    function which maps strings into strings."""

    def __init__(self, function):
        super(FunctionFilter, self).__init__()
        self.function = function

    def write(self, data):
        self.sink.write(self.function(data))


class IndentFilter(ConcreteFilter):

    """Automatically indent a fixed number of spaces after every
    newline."""

    default = ' '

    def __init__(self, indent):
        super(IndentFilter, self).__init__()
        if indent is None:
            self.indent = self.default
        elif isinstance(indent, int):
            self.indent = self.default * indent
        else:
            self.indent = indent

    def write(self, data):
        self.sink.write(data.replace('\n', '\n' + self.indent))


class StringFilter(ConcreteFilter):

    """A filter that takes a lookup table (e.g., string, list, or
    dictionary) and filters any incoming data through it via
    `str.translate`."""

    def __init__(self, table):
        super(StringFilter, self).__init__()
        self.table = table

    def write(self, data):
        self.sink.write(data.translate(self.table))


class BufferedFilter(ConcreteFilter):

    """A buffered filter is one that doesn't modify the source data
    sent to the sink, but instead holds it for a time.  The standard
    variety only sends the data along when it receives a flush
    command."""

    def __init__(self):
        super(BufferedFilter, self).__init__()
        self.buffer = ''

    def write(self, data):
        self.buffer += data

    def flush(self):
        if self.buffer:
            self.sink.write(self.buffer)
        self._flush()


class SizeBufferedFilter(BufferedFilter):

    """A size-buffered filter only in fixed size chunks (excepting the
    final chunk)."""

    def __init__(self, bufferSize=em.Configuration.defaultBuffering):
        super(SizeBufferedFilter, self).__init__()
        assert bufferSize > 0
        self.bufferSize = bufferSize

    def write(self, data):
        BufferedFilter.write(self, data)
        while len(self.buffer) > self.bufferSize:
            chunk, self.buffer = self.buffer[:self.bufferSize], self.buffer[self.bufferSize:]
            self.sink.write(chunk)


class FullyBufferedFilter(BufferedFilter):

    """A maximally-buffered filter only lets its data through on the final
    close.  It ignores flushes."""

    def flush(self): pass

    def close(self):
        if self.buffer:
            BufferedFilter.flush(self)
            self.sink.close()


class DelimitedFilter(BufferedFilter):

    """A delimited filter only lets data through when it sees
    whole lines."""

    def __init__(self, delimiter):
        super(DelimitedFilter, self).__init__()
        self.delimiter = delimiter

    def write(self, data):
        BufferedFilter.write(self, data)
        chunks = self.buffer.split(self.delimiter)
        for chunk in chunks[:-1]:
            self.sink.write(chunk + self.delimiter)
        self.buffer = chunks[-1]


class LineDelimitedFilter(DelimitedFilter):

    """A line-delimited filter only lets data through when it sees
    whole lines."""

    def __init__(self, delimiter=em.NEWLINE_CHAR):
        super(LineDelimitedFilter, self).__init__(delimiter)

    def write(self, data):
        DelimitedFilter.write(self, data)
        chunks = self.buffer.split(self.delimiter)
        for chunk in chunks[:-1]:
            self.sink.write(chunk + self.delimiter)
        self.buffer = chunks[-1]


class DroppingFilter(ConcreteFilter):

    """A filter which drops any chunks that match the provided list of
    chunks to ignore."""

    def __init__(self, droppings):
        super(DroppingFilter, self).__init__()
        self.droppings = droppings

    def write(self, data):
        if data not in self.droppings:
            self.sink.write(data)


class SplittingFilter(Filter):

    """A filter that splits the output stream n ways."""

    def __init__(self, sinks):
        super(SplittingFilter, self).__init__()
        self.sinks = sinks[:]

    def _flush(self):
        for sink in self.sinks:
            sink._flush()

    def write(self, data):
        for sink in self.sinks:
            sink.write(data)

    def flush(self):
        for sink in self.sinks:
            sink.flush()

    def close(self):
        for sink in self.sinks:
            sink.close()

#
# Container
#

Container = em.Plugin # DEPRECATED

#
# AbstractHook, Hook ...
#

class AbstractHook(em.Plugin):

    """An abstract base class for implementing hooks.  All
    hook invocations are not present.  (Use this if you plan
    to systematically handle all calls with something like
    __getattr__.)"""

    pass


class Hook(AbstractHook):

    """The base class for implementing hooks.  All hook
    invocations are present and no-ops."""

    # Override these in a subclass.

    def test(self): pass # used in tests

    # Events

    def atInstallProxy(self, proxy, new): pass
    def atUninstallProxy(self, proxy, done): pass

    def atInstallFinder(self, finder): pass
    def atUninstallFinder(self, finder): pass

    def atStartup(self): pass
    def atReady(self): pass
    def atFinalize(self): pass
    def atShutdown(self): pass
    def atParse(self, scanner, locals): pass
    def atToken(self, token): pass
    def atHandle(self, info, fatal, contexts): pass
    def atInteract(self): pass

    # Contexts

    def pushContext(self, context): pass
    def popContext(self, context): pass
    def setContext(self, context): pass
    def restoreContext(self, context): pass

    # Tokens

    def preLineComment(self, comment): pass

    def preInlineComment(self, comment): pass

    def preWhitespace(self, whitespace): pass

    def preEnable(self, comment): pass
    def preDisable(self, comment): pass

    def prePrefix(self): pass

    def preString(self, string): pass
    def postString(self): pass

    def preBackquote(self, literal): pass
    def postBackquote(self, result): pass

    def preSignificator(self, key, value, stringized): pass
    def postSignificator(self): pass

    def preContextName(self, name): pass
    def postContextName(self, context): pass

    def preContextLine(self, line): pass
    def postContextLine(self, context): pass

    def preExpression(self, pairs, except_, locals): pass
    def postExpression(self, result): pass

    def preSimple(self, code, subtokens, locals): pass
    def postSimple(self, result): pass

    def preInPlace(self, code, locals): pass
    def postInPlace(self, result): pass

    def preStatement(self, code, locals): pass
    def postStatement(self): pass

    def preControl(self, type, rest, locals): pass
    def postControl(self): pass

    def preEscape(self, code): pass
    def postEscape(self): pass

    def preDiacritic(self, code): pass
    def postDiacritic(self): pass

    def preIcon(self, code): pass
    def postIcon(self): pass

    def preEmoji(self, name): pass
    def postEmoji(self): pass

    def preExtension(self, name, contents, depth, locals): pass
    def postExtension(self, result): pass

    def preCustom(self, contents): pass # DEPRECATED
    def postCustom(self, result): pass # DEPRECATED

    # Interpreter actions

    def beforeProcess(self, command, n): pass
    def afterProcess(self): pass

    def beforeInclude(self, file, locals, name): pass
    def afterInclude(self): pass

    def beforeExpand(self, string, locals, name, dispatcher): pass
    def afterExpand(self, result): pass

    def beforeFileLines(self, file, locals, dispatcher): pass
    def afterFileLines(self): pass

    def beforeFileChunks(self, file, bufferSize, locals, dispatcher): pass
    def afterFileChunks(self): pass

    def beforeFileFull(self, file, locals, dispatcher): pass
    def afterFileFull(self): pass

    def beforeImport(self, filename, module, locals, dispatcher): pass
    def afterImport(self): pass

    def beforeString(self, string, locals, dispatcher): pass
    def afterString(self): pass

    def beforeTokens(self, tokens, locals): pass
    def afterTokens(self, result): pass

    def beforeQuote(self, string): pass
    def afterQuote(self, result): pass

    def beforeEscape(self, string, more): pass
    def afterEscape(self, result): pass

    def beforeSignificate(self, key, value, locals): pass
    def afterSignificate(self): pass

    def beforeCallback(self, contents): pass # DEPRECATED
    def afterCallback(self, result): pass # DEPRECATED

    def beforeAtomic(self, name, value, locals): pass
    def afterAtomic(self): pass

    def beforeMulti(self, names, values, locals): pass
    def afterMulti(self): pass

    def beforeClause(self, catch, locals): pass
    def afterClause(self, exception, variable): pass

    def beforeDictionary(self, code, locals): pass
    def afterDictionary(self, result): pass

    def beforeSerialize(self, expression, locals): pass
    def afterSerialize(self): pass

    def beforeDefined(self, name, locals): pass
    def afterDefined(self, result): pass

    def beforeLiteral(self, text, locals): pass
    def afterLiteral(self, result): pass

    def beforeFunctional(self, code, lists, locals): pass
    def afterFunctional(self, result): pass

    def beforeEvaluate(self, expression, locals, replace): pass
    def afterEvaluate(self, result): pass

    def beforeExecute(self, statements, locals): pass
    def afterExecute(self): pass

    def beforeSingle(self, source, locals): pass
    def afterSingle(self, result): pass

    def beforeFinalizer(self, finalizer): pass
    def afterFinalizer(self): pass

#
# Callback
#

class Callback(em.Plugin):

    """A root class for any callback utilities."""

    pass

#
# Finalizer
#

class Finalizer(em.Plugin):

    """A root class for a finalizer."""

    pass

#
# Handler
#

class Handler(em.Plugin):

    """A root class for any error handler."""

    pass

#
# Document
#

class Document(em.Root):

    """A representation of an individual EmPy document, as used by a
    processor."""

    def __init__(self, ident, filename):
        self.ident = ident
        self.filename = filename
        self.significators = {}

#
# Processor
#

class Processor(em.Root):

    """An entity which is capable of processing a hierarchy of EmPy
    files and building a dictionary of document objects associated
    with them describing their significator contents."""

    defaultExtensions = ['.em']

    def __init__(self, config, factory=Document):
        self.config = config
        self.factory = factory
        self.documents = {}

    def identifier(self, pathname, filename): return filename

    def clear(self):
        self.documents = {}

    def scan(self, basename, extensions=None):
        if extensions is None:
            extensions = self.defaultExtensions
        if isinstance(extensions, em.strType):
            extensions = [extensions]
        def _noCriteria(x):
            return True
        def _extensionsCriteria(pathname, extensions=extensions):
            if extensions:
                for extension in extensions:
                    if pathname[-len(extension):] == extension:
                        return True
                return False
            else:
                return True
        self.directory(basename, _noCriteria, _extensionsCriteria, None)
        self.postprocess()

    def postprocess(self):
        pass

    def directory(self, basename, dirCriteria, fileCriteria, depth=None):
        if depth is not None:
            if depth <= 0:
                return
            else:
                depth -= 1
        filenames = os.listdir(basename)
        for filename in filenames:
            pathname = os.path.join(basename, filename)
            if os.path.isdir(pathname):
                if dirCriteria(pathname):
                    self.directory(pathname, dirCriteria, fileCriteria, depth)
            elif os.path.isfile(pathname):
                if fileCriteria(pathname):
                    documentID = self.identifier(pathname, filename)
                    document = self.factory(documentID, pathname)
                    self.file(document, open(pathname))
                    self.documents[documentID] = document

    def file(self, document, file):
        while True:
            line = file.readline()
            if not line:
                break
            self.line(document, line)

    def line(self, document, line):
        significatorRe = re.compile(self.config.significatorReString())
        match = significatorRe.search(line)
        if match:
            key, valueS = match.groups()
            valueS = valueS.strip()
            if valueS:
                value = eval(valueS)
            else:
                value = None
            document.significators[key] = value

#
# Requirement ...
#

class Requirement(em.Root):

    def check(self, details): raise NotImplementedError


class ConjunctiveRequirement(Requirement):

    def __init__(self, requirements):
        self.requirements = requirements

    def check(self, details):
        for requirement in self.requirements:
            if not requirement.check(details):
                return False
        return True


class DisjunctiveRequirement(Requirement):

    def __init__(self, requirements):
        self.requirements = requirements

    def check(self, details):
        for requirement in self.requirements:
            if requirement.check(details):
                return True
        return False


class VersionRequirement(Requirement):

    operators = {
        '=~': lambda x, y: y == x[:min(len(x), len(y))],
        '==': operator.eq,
        '!=': operator.ne,
        '>=': operator.ge,
        '>':  operator.gt,
        '<=': operator.le,
        '<':  operator.lt,
    }

    def __init__(self, operator, version):
        if isinstance(operator, em.strType):
            operator = self.operators[operator]
        self.operator = operator
        if isinstance(version, em.strType):
            version = Details.unpack(version)
        self.version = version

    def check(self, details):
        thisVersion = Details.unpack(details.getPythonVersion())
        return self.operator(thisVersion, self.version)


class ImplementationRequirement(Requirement):

    identity = lambda x: x
    operators = {
        '':  identity,
        '+': identity,
        '-': operator.not_,
    }

    def __init__(self, operator, implementation):
        if isinstance(operator, em.strType):
            operator = self.operators[operator]
        self.operator = operator
        self.implementation = implementation

    def check(self, details):
        thisImplementation = details.getPythonImplementation()
        return self.operator(self.implementation == thisImplementation)

#
# Details
#

class Details(em.Root):

    """Gather details on a running Python system for debugging
    purposes."""

    releaseFilenames = ['/etc/os-release', '/usr/lib/os-release']
    delimiter = '/'
    unsanitizableNames = set(['URL', 'ID'])

    @staticmethod
    def pack(tuple, delimiter='.'):
        """Pack a sequence of ints into a string."""
        return delimiter.join([str(x) for x in tuple])

    @staticmethod
    def unpack(string, delimiter='.'):
        """Unpack a string into a sequence of ints."""
        def convert(x):
            if x.isdigit():
                return int(x)
            else:
                return x
        return tuple([convert(x) for x in string.split(delimiter)])

    def __init__(self, useSanitization=False):
        self.useSanitization = useSanitization
        self.data = {}
        self._system = None
        self._os = None
        self._machine = None
        self._implementation = None
        self._version = None
        self._framework = None
        self._context = None

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        items = list(self.items())
        items.sort()
        return iter(items)

    def empty(self):
        return len(self.data) == 0

    # System/OS information.

    def transformName(self, value, pairs, defaultFormat='[%s]', blank='?'):
        """Find the prefix of the value in the pairs and return
        the resulting name."""
        if not value:
            return blank
        for prefix, name in pairs:
            if value.startswith(prefix):
                return name
        return defaultFormat % value

    def getSystemName(self):
        """Get a nice name for the system this interpreter is
        running on.  Cached."""
        if self._system is None:
            self._system = platform.system()
        assert self._system is not None
        return self._system

    def getOSName(self):
        """Get a nice name for the OS this interpreter is
        running on.  Cached."""
        if self._os is None:
            PAIRS = [
                ('posix', 'POSIX'),
                ('nt', 'NT'),
                ('java', 'Java'),
            ]
            self._os = self.transformName(os.name, PAIRS)
        assert self._os is not None
        return self._os

    # Python information.

    def getPythonImplementation(self):
        """Get the name of this implementation.  Cached."""
        if self._implementation is None:
            DEFAULT = 'CPython'
            try:
                self._implementation = platform.python_implementation()
            except AttributeError:
                # If the module does not contain the function, then it's old
                # enough that we're surely the default.  Fall through.
                pass
            if not self._implementation:
                self._implementation = DEFAULT
        assert self._implementation is not None
        return self._implementation

    def checkPythonImplementation(self, choices):
        """Check the Python implementation against the sequence of
        choices."""
        return self.getPythonImplementation() in choices

    def getPythonVersion(self):
        """Get the version of Python.  Cached."""
        if self._version is None:
            try:
                self._version = platform.python_version()
            except AttributeError:
                # Stackless Python 2.4 raises an internal error when this is
                # called since sys.version is in an unexpected format.
                self._version = self.pack(sys.version_info[:3])
        assert self._version is not None
        return self._version

    def parsePythonVersion(self, version):
        """Parse a Python version into a tuple of ints."""
        if isinstance(version, em.strType):
            version = Details.unpack(version)
        return version

    def checkPythonVersion(self, minimum, maximum=None, closed=True):
        """Check whether or not this Python version is greater than or
        equal to the minimum, and if present, less than or equal to (or
        less than) the maximum."""
        version = sys.version_info[:3]
        minimum = self.parsePythonVersion(minimum)
        if version < minimum:
            return False
        if maximum is not None:
            maximum = self.parsePythonVersion(maximum)
            if len(version) > len(maximum):
                # Trim the version down so it's no more detailed than the
                # maximum.
                version = version[:len(maximum)]
            if closed:
                test = version > maximum
            else:
                test = version >= maximum
            if test:
                return False
        return True

    def getMachineType(self):
        """Get the machine name.  Cached."""
        if self._machine is None:
            self._machine = platform.machine()
            if not self._machine:
                self._machine = '?'
        assert self._machine is not None
        return self._machine

    def getFramework(self):
        """Get the framework and version this interpreter is
        running under as a 2-tuple, or None.  Cached."""
        if self._framework is None:
            implementation = self.getPythonImplementation()
            if implementation == 'PyPy':
                # The PyPy version is in the second line of the version string.
                lines = sys.version.split('\n')
                second = lines[1]
                if second.startswith('['):
                    second = second[1:]
                fields = second.split()
                self._framework = 'PyPy', fields[1]
            elif implementation == 'Jython':
                from java.lang import System
                self._framework = 'JDK', System.getProperty('java.version')
            elif implementation == 'IronPython':
                from System import Environment
                self._framework = '.NET', Environment.Version
            else:
                # CPython
                try:
                    fields = platform.python_compiler().split()
                    pair = fields[:2]
                    rest = fields[3:]
                    self._framework = tuple(pair)
                except AttributeError:
                    # Stackless Python 2.4 raises an internal error when this
                    # is called since sys.version is in an unexpected format.
                    lines = sys.version.splitlines()
                    compiler = lines[-1]
                    if compiler.startswith('[') and compiler.endswith(']'):
                        compiler = compiler[1:-1]
                    self._framework = tuple(compiler.split(' ', 1))
        assert self._framework is not None
        return self._framework

    def getContext(self):
        """Get the context that this interpreter is running in as a
        string.  Note: This is not always possible.  Cached."""
        release = platform.release().lower()
        if self._context is None:
            if 'pyrun_config' in sys.modules:
                self._context = 'PyRun'
            elif 'Stackless' in sys.version:
                fields = sys.version.split(None, 3)
                self._context = fields[1] + '/' + fields[2]
            elif 'ActiveState' in sys.copyright:
                self._context = 'ActiveState'
            elif ('microsoft' in release or 'wsl' in release or
                  'WSL_DISTRO_NAME' in os.environ):
                version = '?'
                # It's WSL.
                if 'wsl2' in release:
                    version = 2
                else:
                    version = 1
                self._context = 'WSL/' + str(version)
            else:
                self._context = ''
            assert self._context is not None
        return self._context

    # Details.

    def sanitize(self, suffix):
        """Normalize this key suffix if necessary."""
        if self.useSanitization:
            if '_' in suffix:
                words = suffix.split('_')
                for i, word in enumerate(words):
                    if i == 0:
                        words[i] = word.lower()
                    elif word in self.unsanitizableNames:
                        words[i] = word
                    else:
                        words[i] = word.capitalize()
                suffix = ''.join(words)
            if suffix.isupper():
                suffix = suffix.lower()
            if suffix[0].isupper():
                suffix = suffix[0].lower() + suffix[1:]
        return suffix

    def key(self, prefix, suffix):
        """Make a key out of this prefix and suffix."""
        if prefix and suffix:
            return prefix + self.delimiter + self.sanitize(suffix)
        elif prefix:
            return prefix
        else: # if suffix:
            return self.normalize(suffix)

    def has(self, prefix, suffix):
        """Is this key present?"""
        return self.key(prefix, suffix) in self.data

    def get(self, prefix, suffix, default=None):
        """Get the value for this key."""
        key = self.key(prefix, suffix)
        if key in self:
            return self.data[key]
        else:
            return default

    def set(self, prefix, suffix, value):
        """Set a key/value pair."""
        self.data[self.key(prefix, suffix)] = value

    def wrap(self, prefix, suffixes, func):
        """Replace an existing entry by passing it through a
        function."""
        if not isinstance(suffixes, tuple):
            suffixes = (suffixes,)
        for suffix in suffixes:
            key = self.key(prefix, suffix)
            if key in self.data:
                self.data[key] = func(self.data[key])

    def accumulate(self, prefix, suffixes, object, name,
                   args=None, attr=None, func=None, kwargs=None, delim='/'):
        """Accumulate the results of a method or attribute in a
        dictionary.  If the argument is None, just return the
        attribute; if not, treat it as a function and call it
        with the args and kwargs.  If attr is not None, access
        that named attribute on the result before returning.  If
        func is not None, pass the result through the function
        before storing."""
        assert name is not None
        if prefix is None:
            prefix = ''
        result = getattr(object, name, None)
        if result is None:
            return
        error = None
        if not isinstance(suffixes, tuple):
            suffixes = (suffixes,)
        try:
            if args is not None:
                # It's a function call.
                if kwargs is None:
                    kwargs = {}
                result = result(*args, **kwargs)
            if result is None:
                return
            if not isinstance(result, tuple):
                # If it's not a tuple, make it one for uniformity.
                result = (result,)
            if len(suffixes) == 1 and len(result) > 1:
                # If we have only one key but multiple values, then it's a
                # single key with a value that's a tuple.  Join up the tuple
                # with the delimiter.
                result = (delim.join(result),)
            for suffix, value in zip(suffixes, result):
                # Now iterate over the suffixes/values:
                if suffix is None:
                    # Skip suffixes that are None.
                    continue
                if attr is not None:
                    # We want an attribute of it.
                    value = getattr(value, attr)
                if func is not None:
                    # If there's a function wrapper, call it.
                    value = func(value)
                assert suffix not in self.data
                self.set(prefix, suffix, value)
        except:
            # Some weird things can happen when calling/accessing the result:
            #
            # - Jython 2.5 raises an internal AttributeError when accessing
            #   platform.architecture.
            # - IronPython 2.7 has not implemented platform.architecture.
            # - Some older versions of IronPython 2.7 raise an ImportError when
            #   calling platform.platform which is then bizarrely reported as a
            #   TypeError with no traceback by the executive.
            #
            # Set the appropriate values with a question mark and the name of
            # the exception which occurred to make it clear what happened.
            type, error, traceback = sys.exc_info()
            value = '?' + error.__class__.__name__
            for suffix in suffixes:
                self.set(prefix, suffix, value)

    # Requirements.

    def parseRequirement(self, string):
        DISJUNCTION = ';'
        VERSION_OPERATORS = '><=!~'
        IMPLEMENTATION_OPERATORS = '-'
        strings = string.split(';')
        results = []
        for string in strings:
            assert string
            if string[0] in VERSION_OPERATORS:
                # It's a version requirement.
                i = 0
                while i < len(string) and string[i] in VERSION_OPERATORS:
                    i += 1
                op, ver = string[:i], string[i:].strip()
                results.append(VersionRequirement(op, ver))
            else:
                # It's an implementation requirement.
                i = 0
                while i < len(string) and string[i] in IMPLEMENTATION_OPERATORS:
                    i += 1
                op, impl = string[:i], string[i:].strip()
                results.append(ImplementationRequirement(op, impl))
        if len(results) == 1:
            return results[0]
        else:
            return DisjunctiveRequirement(results)

    def parseRequirements(self, filename):
        requirements = []
        file = open(filename)
        try:
            for line in file.readlines():
                line = line.strip()
                if not line:
                    continue
                elif line.startswith('#'):
                    continue
                requirement = self.parseRequirement(line)
                requirements.append(requirement)
        finally:
            file.close()
        return requirements

    def checkRequirements(self, requirements):
        if isinstance(requirements, em.strType):
            requirements = self.parseRequirements(requirements)
        for requirement in requirements:
            if not requirement.check(self):
                return False
        return True

    # Details.

    def getBasicDetails(self):
        """Collect basic details."""
        self.accumulate('basic', 'system', self, 'getSystemName', ())
        self.accumulate('basic', 'os', self, 'getOSName', ())
        self.accumulate('basic', 'machine', self, 'getMachineType', ())
        self.accumulate('basic', 'implementation', self, 'getPythonImplementation', ())
        self.accumulate('basic', 'version', self, 'getPythonVersion', ())
        self.accumulate('basic/framework', ('name', 'version'), self, 'getFramework', ())
        self.accumulate('basic', 'context', self, 'getContext', ())

    def getPythonDetails(self):
        """Collect details about this interpreter.  Use the
        platform module if available when the flag is set to
        true."""
        self.accumulate('python', 'implementation', self, 'getPythonImplementation', ())
        self.accumulate('python', 'version', self, 'getPythonVersion', ())
        self.accumulate('python', 'branch', platform, 'python_branch', ())
        self.accumulate('python/build', ('name', 'date'), platform, 'python_build', ())
        self.accumulate('python', 'compiler', platform, 'python_compiler', ())
        self.accumulate('python', 'revision', platform, 'python_revision', ())

    def getSystemDetails(self):
        """Collect details about the system settings on this
        interpreter."""
        self.accumulate('system/api', 'flags', sys, 'abiflags')
        self.accumulate('system/api', 'version', sys, 'api_version')
        self.accumulate('system', 'byteorder', sys, 'byteorder')
        self.accumulate('system', 'copyright', sys, 'copyright', func=repr)
        self.accumulate('system/filesystem', 'encoding', sys, 'getfilesystemencoding', ())
        self.accumulate('system/filesystem', 'errors', sys, 'getfilesystemencodeerrors', ())
        self.accumulate('system/float', ('max', 'max_exp', None, 'min', 'min_exp', None, 'dig', 'mant_dig', 'epsilon', 'radix', 'rounds'), sys, 'float_info')
        self.accumulate('system/hash', ('width', 'modulus', 'inf', 'nan', 'imag', 'algorithm', 'bits', 'seed_bits'), sys, 'hash_info')
        self.wrap('system/hash', 'modulus', hex)
        self.accumulate('system/int', ('bits', 'sizeof', 'default_max', 'check_threshold'), sys, 'int_info')
        self.accumulate('system/path', 'executable', sys, 'executable')
        self.accumulate('system/path', 'library', sys, 'platlibdir')
        self.accumulate('system/path', 'prefix', sys, 'prefix')
        self.accumulate('system', 'platform', sys, 'platform')
        self.accumulate('system', 'size/max', sys, 'maxsize', func=hex)
        self.set('system', 'unicode/build', ['wide', 'narrow'][em.narrow])
        self.accumulate('system', 'unicode/max', sys, 'maxunicode', func=hex)
        self.accumulate('system/version', 'hex', sys, 'hexversion', func=hex)
        self.accumulate('system/version', 'str', sys, 'version', func=repr)

    def getPlatformDetails(self):
        """Collect details about this platform.  Use the
        platform module if available when the flag is set to true."""
        self.accumulate('platform/architecture', ('bits', 'linkage'), platform, 'architecture', ())
        self.accumulate('platform', 'machine', platform, 'machine', ())
        self.accumulate('platform', 'name', platform, 'platform', ())
        self.accumulate('platform', 'node', platform, 'node', ())
        self.accumulate('platform', 'processor', platform, 'processor', ())
        self.accumulate('platform', 'release', platform, 'release', ())
        self.accumulate('platform', 'system', platform, 'system', ())
        self.accumulate('platform/thread', ('implementation', 'lock', 'version'), sys, 'thread_info')
        self.accumulate('platform', 'version', platform, 'version', ())

    def getReleaseDetails(self, system=None):
        """Collect details about the given system release, or the
        running one if None."""
        if system is None:
            system = self.getSystemName()
        methodName = 'getReleaseDetails_' + system
        method = getattr(self, methodName, None)
        if method is not None:
            method()

    # Release details specializations.

    def getReleaseDetails_Linux(self):
        """Collect details about the Linux release."""
        PREFIX = 'release/linux'
        file = None
        for filename in self.releaseFilenames:
            try:
                file = open(filename, 'r')
                break
            except IOError:
                pass
        if file is None:
            return self.data
        try:
            while True:
                line = file.readline()
                if not line:
                    break
                if line.endswith('\n'):
                    line = line[:-1]
                key, value = line.split('=', 1)
                if value.startswith('"') and value.endswith('"'):
                    value = value[1:-1]
                self.set(PREFIX, key, value)
        finally:
            file.close()

    def getReleaseDetails_Darwin(self):
        """Collect details about the Darwin release."""
        PREFIX = 'release/darwin'
        PROGRAM = 'sw_vers'
        FILENAME = '/tmp/sw_vers_%d.out' % os.getpid()
        waitStatus = os.system('%s > %s 2> /dev/null' % (PROGRAM, FILENAME))
        exitCode = waitStatus >> 8
        if exitCode != 0:
            return
        file = None
        try:
            try:
                file = open(FILENAME, 'r')
                for line in file:
                    if line.endswith('\n'):
                        line = line[:-1]
                    key, value = line.split(':', 1)
                    key = key.strip()
                    value = value.strip()
                    self.set(PREFIX, key, value)
            except OSError:
                pass
        finally:
            if file is not None:
                file.close()
            try:
                os.remove(FILENAME)
            except OSError:
                pass

    def getReleaseDetails_Windows(self):
        """Collect details about the Windows release."""
        PREFIX = 'release/windows'
        PRODUCT_TYPES = {
            1: 'VER_NT_WORKSTATION',
            2: 'VER_NT_DOMAIN_CONTROLLER',
            3: 'VER_NT_SERVER',
        }
        self.accumulate(PREFIX, 'dllhandle', sys, 'dllhandle', func=hex)
        self.accumulate(PREFIX, 'registry', sys, 'winver')
        self.accumulate(PREFIX, ('major', 'minor', 'build', 'platform', 'service_pack'), sys, 'getwindowsversion', ())
        self.accumulate(PREFIX, 'suite_mask', sys, 'getwindowsversion', (), attr='suite_mask', func=hex)
        self.accumulate(PREFIX, 'product_type', sys, 'getwindowsversion', (), attr='product_type', func=lambda x: PRODUCT_TYPES.get(x, x))
        self.accumulate(PREFIX, 'platform_version', sys, 'getwindowsversion', (), attr='platform_version', func=lambda x: '.'.join([em.nativeStr(e) for e in x]))

    # Summary.

    def classify(self, version):
        """Classify a release by version string."""
        PAIRS = [
            ('a', 'alpha'),
            ('b', 'beta'),
            ('d', 'development'),
            ('g', 'gamma'),
            ('h', 'adhoc'),
            ('r', 'revision'),
            ('RC', 'release candidate'),
            ('v', 'version'),
            ('x', 'development'),
        ]
        # First, normalize it.
        if '-' in version and '.' not in version:
            version = version.replace('-', '.')
        major = version.split('.')[0]
        if len(major) == 4 and major.isdigit():
            # If the major version looks like a year, it's a preview version.
            return 'preview'
        for key, name in PAIRS:
            if key in version:
                return name
        return None

    def collect(self, level):
        """Collect details."""
        if level >= em.Version.BASIC:
            self.getBasicDetails()
        if level >= em.Version.PYTHON:
            self.getPythonDetails()
        if level >= em.Version.SYSTEM:
            self.getSystemDetails()
        if level >= em.Version.PLATFORM:
            self.getPlatformDetails()
        if level >= em.Version.RELEASE:
            self.getReleaseDetails()

    def show(self, level=em.Version.INFO, prelim="", postlim="", file=None):
        """Show details."""
        if file is None:
            file = sys.stdout
        write = file.write
        if prelim:
            write(prelim)
        write("%s version %s" % (em.__project__, em.__version__))
        classification = self.classify(em.__version__)
        if classification:
            write(" [%s]" % classification)
        if level >= em.Version.INFO:
            write(", in %s/%s" % (
                self.getPythonImplementation(), self.getPythonVersion()))
            context = self.getContext()
            if context:
                write(", as %s" % context)
            write(", on %s (%s)" % (
                self.getSystemName(), self.getOSName()))
            write(", with %s" % self.getMachineType())
            framework = self.getFramework()
            if framework:
                write(", under %s/%s" % framework)
            if em.compat:
                write(" (%s)" % ', '.join(em.compat))
        if postlim:
            write(postlim)
        if self.empty():
            self.collect(level)
        if level >= em.Version.DATA:
            if self.data:
                write("Details:\n")
                items = list(self.data.items())
                items.sort()
                for key, value in items:
                    if value is None or value == '':
                        value = '--'
                    write("- %s: %s\n" % (key, value))
            else:
                write("(No details available.)\n")

#
# main
#

def main():
    if len(sys.argv) > 1:
        arg = sys.argv[1]
        try:
            level = int(arg)
        except ValueError:
            level = getattr(em.Version, arg)
    else:
        level = em.Version.ALL
    details = Details()
    details.show(level, prelim="Welcome to ", postlim=".\n")

if __name__ == '__main__': main()



#!/usr/bin/env python3

"""
Help subsystem for EmPy.
"""

#
# imports
#

import sys

import em

#
# Entry ...
#

class Entry(em.Root):

    """An entry in one of the usage tables.  This exists to allow
    optional annotations or processing for each entry."""

    def __init__(self, raw, right,
                 var=None, val=None, env=None, arg=None,
                 ord=None, ex=None, fun=None):
        """Create an entry for a help topic.  Arguments are:

        - raw: The left side raw value: str | list[str]
        - right: The right side base value: str
        - var: The configuration variable, if any: Optional[str]
        - val: The boolean value of the variable, if any: Optional[bool]
        - env: The environment variable name, if any: Optional[str]
        - arg: The command line option arguments, if any: Optional[str]
        - ord: The Unicode code point value(s), if any: Optional[int |
          list[int]]
        - ex: A practical example, if any: Optional[str]
        - fun: An optional function to format the resulting string:
          Optional[Callable]
        - sim: Whether or not the option entry is simple (-x): bool
        """
        self.raw = raw
        self.left = self.format(raw)
        self.right = right
        self.var = var
        self.val = val
        self.env = env
        self.arg = arg
        if isinstance(ord, int):
            ord = [ord]
        self.ords = ord
        self.ex = ex
        self.fun = fun
        if isinstance(self.raw, list):
            self.sim = self.raw[0] != '  '
        else:
            self.sim = False

    def __str__(self):
        if self.ords is None:
            ords = '--'
        else:
            ords = ' '.join([hex(x) for x in self.ords])
        return '%s [%s]' % (self.left, ords)

    def key(self):
        return (self.ords, self.left)

    def format(self, raw):
        """Format the raw list.  Override in subclasses."""
        if isinstance(raw, list):
            return ' '.join(raw)
        else:
            return raw


class NameEntry(Entry):

    """An entry tailored to a name."""

    def key(self):
        return self.left


class OptionEntry(Entry):

    """An entry tailored to an option."""

    def format(self, raw):
        if isinstance(raw, list):
            assert len(raw) > 0
            if raw[0].startswith('--'):
                raw.insert(0, '  ')
            return ' '.join(raw)
        else:
            return raw


class ControlEntry(Entry):

    """An entry tailored to a control (named escape)."""

    def key(self):
        name = self.left
        if name.isdigit():
            name = '_' + name
        return (self.ords, name)

#
# Section ...
#

class Section(em.Root):

    """A section in one of the usage tables."""

    def __init__(self, topic, description, header):
        self.topic = topic
        self.description = description
        self.header = header

    def __str__(self):
        return self.topic

    def ok(self):
        return True

    def show(self, usage, useSeparator=True):
        if not self.ok():
            return
        if useSeparator:
            usage.separator()
        usage.write("%s:\n" % self.topic.upper())
        usage.write(self.header)
        self.block(usage)

    def block(self, usage):
        pass


class TableSection(Section):

    """A section with a block that consists of a flat table of
    entries."""

    def __init__(self, topic, description, header, entries):
        super(TableSection, self).__init__(topic, description, header)
        self.entries = entries

    def ok(self):
        return bool(self.entries)

    def block(self, usage):
        usage.table(self.entries)


class ConfigSection(TableSection):

    """A section with a block that consists of each configuration
    variable."""

    def __init__(self, topic, description, header, config):
        super(ConfigSection, self).__init__(topic, description, header, [])
        for left in config._names:
            right = config._descriptions[left]
            fun = config._functions[left]
            entry = Entry(left, right, var=left, fun=fun)
            self.entries.append(entry)


class MappingSection(TableSection):

    """A section with a block that consists of a mapping type which is
    transformed into a table of entries."""

    def __init__(self, topic, description, header, mapping, factory=Entry):
        super(MappingSection, self).__init__(topic, description, header, [])
        for key, value in mapping.items():
            if value is None:
                continue
            # Value can be an ordinal, a string, or a 2-tuple of (ordinal,
            # name).
            if isinstance(value, tuple):
                ordinal, right = value
            else:
                ordinal = None
                right = value
            if isinstance(ordinal, (em.bytes, em.str)):
                ordinal = ord(ordinal)
            if isinstance(right, int):
                right = em.chr(right)
            elif isinstance(right, list):
                right = ''.join([em.chr(x) for x in right])
            entry = factory(key, right, ord=ordinal)
            self.entries.append(entry)
        self.entries.sort(key=lambda x: x.key())


class HelpSection(TableSection):

    """A section with a block that consists of the help topics."""

    def __init__(self, topic, description, header, usage):
        super(HelpSection, self).__init__(topic, description, header, [])
        self.usage = usage

    def refresh(self):
        """Refresh the list of topics.  At creation time it won't
        contain itself."""
        payload = self.usage.payload
        all = Usage.aliases['all']
        for topic in all:
            if topic in payload:
                section = payload[topic]
                if section.ok():
                    self.entries.append(
                        Entry(section.topic, section.description))

#
# Usage
#

class Usage(em.Root):

    """A utility class to print usage and extended help."""

    aliases = {
        # Groups
        'default': ['usage', 'options', 'markup', 'hints', 'topics'],
        'more': ['usage', 'options', 'markup', 'escapes', 'environ', 'hints',
                 'topics'],
        'all': ['usage', 'options', 'simple',
                'markup', 'escapes', 'environ', 'pseudo',
                'constructor', 'variables', 'methods', 'hooks', 'named',
                'diacritics', 'icons', 'emojis', 'hints', 'topics'],
        # Aliases
        'accents': ['diacritics'],
        'config': ['variables', 'methods'],
        'control': ['named'],
        'controls': ['named'],
        'ctor': ['constructor'],
        'diacritic': ['diacritics'],
        'emoji': ['emojis'],
        'escape': ['escapes'],
        'flags': ['options'],
        'hint': ['hints'],
        'hook': ['hooks'],
        'icon': ['icons'],
        'markups': ['markup'],
        'method': ['methods'],
        'option': ['options'],
        'optional': ['emojis'],
        'switches': ['options'],
        'topic': ['topics'],
        'variable': ['variables'],
    }

    defaultWidth = 6
    columns = None # 81
    dottedCircle = '\N{DOTTED CIRCLE}'

    def __init__(self, config=None, file=None):
        if config is None:
            config = em.Configuration()
        self.config = config
        if file is None:
            file = sys.stdout
        self.file = file
        self.payload = {}
        prepare(self)
        self.fix()

    def format(self, width):
        return '  %%-%ds  %%s\n' % max(width, Usage.defaultWidth)

    def transform(self, topics):
        """Transform a list of possible topics into a list of valid
        topics."""
        if topics is None:
            topics = 'default'
        results = []
        if not isinstance(topics, list):
            topics = [topics]
        for topic in topics:
            topic = topic.lower()
            if topic in Usage.aliases:
                results.extend(Usage.aliases[topic])
            else:
                results.append(topic)
        return results

    def add(self, section):
        """Add a section to the payload."""
        assert section.topic not in self.payload
        self.payload[section.topic] = section

    def fix(self):
        """Fix the payload."""
        self.payload['topics'].refresh()

    def separator(self):
        """Write a separator."""
        self.write("\n")

    def write(self, string):
        """Write a string (with the correct prefix substitution)."""
        if not self.config.hasDefaultPrefix():
            string = string.replace(self.config.defaultPrefix,
                                    self.config.prefix)
        self.file.write(string)

    def flush(self):
        """Flush the underlying stream."""
        self.file.flush()

    def scan(self, table):
        """Scan a table to find the minimum column width."""
        result = 0
        for entry in table:
            left = entry.left
            if len(left) > result:
                result = len(left)
        return result

    def preview(self, ords):
        """Return a preview of the (non-text) character ordinals,
        or None."""
        if ords is None:
            return
        if em.major < 3 or not sys.stdout.encoding.lower().startswith('utf'):
            # Don't bother if this isn't Python 3.x or the encoding isn't
            # UTF-8.
            return
        firstOrd = ords[0]
        chars = ''.join([chr(x) for x in ords])
        while not chars.isprintable():
            # Some emoji blocks report as unprintable in various versions of
            # Python.  (This is implemented as a while loop with breaks ending
            # in a return to make it easier to edit.)
            if firstOrd >= 0x2300 and firstOrd < 0x2400:
                break
            if firstOrd >= 0x2600 and firstOrd < 0x2c00:
                break
            if firstOrd >= 0x1f000 and firstOrd < 0x1fb00:
                break
            return
        if chars.isspace():
            return
        if firstOrd == 0x034f:
            # Combining grapheme joiner.
            return
        if firstOrd >= 0xfe00 and firstOrd < 0xfe10:
            # Variation selectors 1-16.
            return
        if firstOrd == 0xfffc:
            # Object replacement character.
            return
        if firstOrd >= 0xe0100 and firstOrd < 0xe01f0:
            # Variation selectors 17-256.
            return
        if firstOrd >= 0x0300 and firstOrd < 0x0370:
            # Combining diacritical marks.
            chars = self.dottedCircle + chars
        if firstOrd >= 0x1ab0 and firstOrd < 0x1b00:
            # Combining diacritical marks extended.
            chars = self.dottedCircle + chars
        if firstOrd >= 0x1dc0 and firstOrd < 0x1e00:
            # Combining diacritical marks supplemental.
            chars = self.dottedCircle + chars
        if firstOrd >= 0x20d0 and firstOrd < 0x2100:
            # Combining diacritical marks for symbols.
            chars = self.dottedCircle + chars
        return chars

    def entry(self, entry, format):
        """Print a table entry."""
        def _identity(x): return x
        left = entry.left
        lines = entry.right.split('\n')
        fun = entry.fun or _identity
        i = 0
        for line in lines:
            last = i == len(lines) - 1
            fragments = []
            if entry.ords is not None:
                first = True
                for x in entry.ords:
                    if first:
                        first = False
                    else:
                        fragments.append(' ')
                    fragments.append("U+%04X" % x)
                fragments.append("; ")
            fragments.append(line)
            if last:
                if entry.arg is not None:
                    fragments.append(": %s" % fun(entry.arg))
                if entry.var is not None and entry.val is None:
                    value = fun(getattr(self.config, entry.var))
                    if isinstance(value, dict):
                        value = "{%d}" % len(value)
                    fragments.append(" [%s]" % str(value))
                if entry.ex is not None:
                    fragments.append(" (e.g., %s)" % fun(entry.ex))
                preview = self.preview(entry.ords)
                if preview:
                    fragments.append(': ')
                    fragments.append(preview)
            right = ''.join(fragments).strip()
            line = format % (left, right)
            assert self.columns is None or len(line) < self.columns, line
            self.write(line)
            left = ''
            i += 1

    def table(self, table):
        """Print a table at the current level with an optional header."""
        width = self.scan(table)
        format = self.format(width)
        for entry in table:
            self.entry(entry, format)

    def hello(self):
        self.write("Welcome to %s version %s.\n" % (
            em.__project__, em.__version__))

    def show(self, topics=None, useSeparator=True):
        """Show usage."""
        if topics is None:
            topics = 'default'
        topics = self.transform(topics)
        flag = False
        for topic in topics:
            topic = topic.lower()
            if topic in self.payload:
                section = self.payload[topic]
                if section.ok():
                    section.show(self, useSeparator)
                    flag = True
            else:
                self.write("*** Unknown usage topic: %s\n" % topic)
        if not flag:
            self.write("*** No active topics found\n")

#
# functions
#

def prepare(usage, executable=None):
    """Lazily initialize the usage data and tables only if they're
    actually needed.  This is a standalone function so there's less
    misleading indentation.  Idempotent."""
    if executable is None:
        executable = sys.argv[0]
    payload = usage.payload
    if payload:
        return payload
    E = Entry
    OE = OptionEntry

    usage.add(Section(
        'usage',
        "Basic command line usage",
        """\
%s [<options>] [<filename, or `-` for stdin> [<argument>...]]
  - Options begin with `-` or `--`
  - Specify a filename (and arguments) to process that document as input
  - Specify `-` (and arguments) to process stdin with standard buffering
  - Specify no filename to enter interactive mode with line buffering
  - Specify `--` to stop processing options
""" % executable))

    usage.add(Section(
        'hints',
        "Usage hints",
        """\
Whitespace immediately inside parentheses of `@(...)` are ignored.  Whitespace
immediately inside braces of `@{...}` are ignored, unless ... spans multiple
lines.  Use `@{ ... }@` to suppress newline following second `@`.  Simple
expressions ignore trailing punctuation; `@x.` means `@(x).`, not a parse
error.  A `#!` at the start of a file is treated as a comment.  The full
documentation is available at <%s>.
""" % em.__url__))

    usage.add(TableSection(
        'options',
        "Command line options",
        """\
Valid command line options (defaults in brackets):
""", [
OE(["-V", "--version"], "Print version"),
OE(["-W", "--info"], "Print version and system information"),
OE(["-Z", "--details"], "Print extensive system and platform details"),
OE(["-h", "--help"], "Print help; more -h options for more help"),
OE(["-H", "--topics=TOPICS"], "Print usage for comma-separated topics"),
OE(["-v", "--verbose"], "Show verbose debugging while processing"),
OE(["-p", "--prefix=CHAR"], "Choose desired prefix", var='prefix', env=em.PREFIX_ENV),
OE(["--no-prefix"], "Do not do any markup processing at all"),
OE(["-q", "--no-output"], "Suppress all output"),
OE(["-m", "--pseudomodule=NAME"], "Change internal pseudomodule name", var='pseudomoduleName', env=em.PSEUDO_ENV),
OE(["-f", "--flatten"], "Flatten pseudomodule members at start", var='doFlatten', env=em.FLATTEN_ENV, val=True),
OE(["-k", "--keep-going"], "Don't exit on errors; continue", var='exitOnError', val=False),
OE(["-e", "--ignore-errors"], "Ignore errors completely", var='ignoreErrors', val=True),
OE(["-r", "--raw-errors"], "Show full tracebacks of Python errors", var='rawErrors', env=em.RAW_ERRORS_ENV, val=True),
OE(["-s", "--brief-errors"], "Don't show attributes when formatting errors", var='verboseErrors', val=False),
OE(["--verbose-errors"], "Show attributes when formatting errors (default)", var='verboseErrors', val=True),
OE(["-i", "--interactive"], "Enter interactive mode after processing", var='goInteractive', env=em.INTERACTIVE_ENV, val=True),
OE(["-d", "--delete-on-error"], "Delete output file on error\n"
                                "(-o or -a needed)", var='deleteOnError', env=em.DELETE_ON_ERROR_ENV),
OE(["-n", "--no-proxy"], "Do not override sys.stdout with proxy", var='useProxy', env=em.NO_PROXY_ENV),
OE(["--config=STATEMENTS"], "Do configuration variable assignments"),
OE(["-c", "--config-file=FILENAME"], "Load configuration resource path(s)", env=em.CONFIG_ENV),
OE(["--config-variable=NAME"], "Configuration variable name while loading", var='configVariableName'),
OE(["-C", "--ignore-missing-config"], "Don't raise for missing configuration files", var='missingConfigIsError', val=False),
OE(["-o", "--output=FILENAME"], "Specify file for output as write"),
OE(["-a", "--append=FILENAME"], "Specify file for output as append"),
OE(["-O", "--output-binary=FILENAME"], "Specify file for binary output as write"),
OE(["-A", "--append-binary=FILENAME"], "Specify file for binary output as append"),
OE(["--output-mode=MODE"], "Explicitly specify the mode for output"),
OE(["--input-mode=MODE"], "Explicitly specify the mode for input"),
OE(["-b", "--buffering"], "Specify buffering strategy for files:\n"
                          "none (= 0), line (= 1), full (= -1), default,\n"
                          "or another value for fixed", var='buffering', env=em.BUFFERING_ENV),
OE(["--default-buffering"], "Specify default buffering for files (default)"),
OE(["-N", "--no-buffering"], "Specify no buffering for reads on files"),
OE(["-L", "--line-buffering"], "Specify line buffering for files"),
OE(["-B", "--full-buffering"], "Specify full buffering for files"),
OE(["-I", "--import=MODULES"], "Import Python modules before processing"),
OE(["-D", "--define=DEFN"], "Execute Python assignment statement"),
OE(["-S", "--string=STR"], "Execute string literal assignment"),
OE(["-P", "--preprocess=FILENAME"], "Interpret EmPy file before main document"),
OE(["-Q", "--postprocess=FILENAME"], "Interpret EmPy file after main document"),
OE(["-E", "--execute=STATEMENT"], "Execute Python statement before main document"),
OE(["-K", "--postexecute=STATEMENT"], "Execute Python statement after main document"),
OE(["-F", "--file=FILENAME"], "Execute Python file before main document"),
OE(["-G", "--postfile=FILENAME"], "Execute Python file after main document"),
OE(["-X", "--expand=MARKUP"], "Expand EmPy markup before main document"),
OE(["-Y", "--postexpand=MARKUP"], "Expand EmPy markup after main document"),
OE(["--preinitializer=FILENAME"], "Execute Python file before interpreter"),
OE(["--postinitializer=FILENAME"], "Execute Python file after interpreter"),
OE(["-w", "--pause-at-end"], "Prompt at the end of processing", var='pauseAtEnd'),
OE(["-l", "--relative-path"], "Add path of EmPy script to sys.path", var='relativePath'),
OE(["--replace-newlines"], "Replace expression newlines with spaces (default)", var='replaceNewlines', val=True),
OE(["--no-replace-newlines"], "Don't replace expression newlines with spaces", var='replaceNewlines', val=False),
OE(["--ignore-bangpaths"], "Treat bangpaths as comments (default)", var='ignoreBangpaths', val=True),
OE(["--no-ignore-bangpaths"], "Don't treat bangpaths as comments", var='ignoreBangpaths', val=False),
OE(["--expand-user"], "Expand user constructions (~user) in paths (default)", var='expandUserConstructions', val=True),
OE(["--no-expand-user"], "Don't expand user constructions in paths", var='expandUserConstructions', val=False),
OE(["--auto-validate-icons"], "Auto-validate icons on first use (default)", var='autoValidateIcons', val=True),
OE(["--no-auto-validate-icons"], "Don't auto-validate icons on first use", var='autoValidateIcons', val=False),
OE(["--none-symbol"], "String to write when expanding None", var='noneSymbol'),
OE(["--no-none-symbol"], "Write nothing when expanding None (default)"),
OE(["--starting-line"], "Line number to start with", var='startingLine'),
OE(["--starting-column"], "Column number to start with", var='startingColumn'),
OE(["--emoji-modules"], "Comma-separated list of emoji modules to try\n", var='emojiModuleNames', fun=lambda x: ','.join(x)),
OE(["--no-emoji-modules"], "Only use unicodedata for emoji markup"),
OE(["--disable-emoji-modules"], "Disable emoji modules; use emojis dictionary"),
OE(["--ignore-emoji-not-found"], "Emoji not found is not an error", var='emojiNotFoundIsError', val=False),
OE(["-u", "--binary", "--unicode"], "Read input file as binary?\n"
                                    "(enables Unicode support in Python 2.x)", var='useBinary', env=em.BINARY_ENV),
OE(["-x", "--encoding=E"], "Set both input and output Unicode encodings"),
OE(["--input-encoding=E"], "Set input Unicode encoding", var='inputEncoding', env=em.INPUT_ENCODING_ENV, fun=lambda x: x == 'utf_8' and 'utf-8' or x),
OE(["--output-encoding=E"], "Set output Unicode encoding", var='outputEncoding', env=em.OUTPUT_ENCODING_ENV, fun=lambda x: x == 'utf_8' and 'utf-8' or x),
OE(["-y", "--errors=E"], "Set both input and output Unicode error handler"),
OE(["--input-errors=E"], "Set input Unicode error handler", var='inputErrors', env=em.INPUT_ERRORS_ENV),
OE(["--output-errors=E"], "Set output Unicode error handler", var='outputErrors', env=em.OUTPUT_ERRORS_ENV),
OE(["-z", "--normalization-form=F"], "Specify Unicode normalization form", var='normalizationForm'),
OE(["--auto-play-diversions"], "Autoplay diversions on exit (default)", var='autoPlayDiversions', val=True),
OE(["--no-auto-play-diversions"], "Don't autoplay diversions on exit", var='autoPlayDiversions', val=False),
OE(["--check-variables"], "Validate configuration variables (default)", var='checkVariables', val=True),
OE(["--no-check-variables"], "Don't validate configuration variables", var='checkVariables', val=False),
OE(["--path-separator"], "Path separator for configuration paths", var='pathSeparator'),
OE(["--enable-modules"], "Enable EmPy module support (default)", var='supportModules', val=True),
OE(["-g", "--disable-modules"], "Disable EmPy module support", var='supportModules', val=False),
OE(["--module-extension"], "Filename extension for EmPy modules", var='moduleExtension'),
OE(["--module-finder-index"], "Index of module finder in meta path", var='moduleFinderIndex'),
OE(["--enable-import-output"], "Enable output during import (default)", var='enableImportOutput', val=True),
OE(["-j", "--disable-import-output"], "Disable output during import", var='enableImportOutput', val=False),
OE(["--context-format"], "Context format", var='contextFormat'),
OE(["--success-code=N"], "Exit code to return on script success", var='successCode'),
OE(["--failure-code=N"], "Exit code to return on script failure", var='failureCode'),
OE(["--unknown-code=N"], "Exit code to return on bad configuration", var='unknownCode'),
]))

    usage.add(TableSection(
        'simple',
        "Simple (one-letter) command line options",
        """\
Short (single letter) command line options:
""",
        sorted([x for x in usage.payload['options'].entries if x.sim],
                           key=lambda x: x.raw[0].swapcase())))

    usage.add(TableSection(
        'markup',
        "Markup syntax",
        """\
The following markups are supported (NL means newline; WS means whitespace):
""", [
E("@# ... NL", "Line comment; consume everything including newline"),
E("@* ... *", "Inline comment; consume everything inside"),
E("@ WS", "Ignore whitespace; line continuation"),
E("@- ... NL", "Disable output; consume up to trailing newline"),
E("@+ ... NL", "Enable output; consume up to trailing newline"),
E("@@", "Literal @; @ is escaped (duplicated prefix)"),
E("@' STRING '", "Replace with string literal contents"),
E("@\" STRING \"", "Replace with string literal contents"),
E("@''' STRING ... '''", "Replace with multiline string literal contents"),
E("@\"\"\" STRING ... \"\"\"", "Replace with multiline string literal contents"),
E("@` LITERAL `", "Write everything inside literally (no expansion)"),
E("@( EXPRESSION )", "Evaluate expression and substitute with str"),
E("@( TEST ? THEN )", "If test, evaluate then, otherwise do nothing"),
E("@( TEST ? THEN ! ELSE )", "If test, evaluate then, otherwise evaluate else;\n"
                             "can be chained with repeated test/then/[else]"),
E("@( TRY $ EXCEPT )", "Evaluate try expression, or except if it raises"),
E("@ SIMPLE_EXPRESSION", "Evaluate simple expression and substitute\n", ex="@x, @x.y, @f(a, b), @l[i], @f{...}, etc."),
E("@$ EXPRESSION $ [DUMMY] $", "Evaluates to @$ EXPRESSION $ EXPANSION $"),
E("@{ STATEMENTS }", "Statements are executed for side effects"),
E("@[ CONTROL ]", "Control markups: if E; elif E; for N in E;\n"
                  "while E; dowhile E; try; except E as N; finally;\n"
                  "continue; break; else; with E as N; match E;\n"
                  "case E; defined N; def F(...); end X"),
E("@\\ ESCAPE_CODE", "A C-style escape sequence"),
E("@\\^{ NAMED_ESCAPE }", "A named escape sequence", ex="ESC for escape"),
E("@^ CHAR DIACRITIC(S)", "A two-part diacritic sequence\n", ex="e' for an e with acute accent"),
E("@| ICON", "A custom icon sequence\n", ex=":) for a smiley face emoji"),
E("@: EMOJI :", "Lookup emoji by name"),
E("@% KEY NL", "Significator form of __KEY__ = None"),
E("@% KEY WS VALUE NL", "Significator form of __KEY__ = VALUE"),
E("@%! KEY NL", "Significator form of __KEY__ = \"\""),
E("@%! KEY WS STRING NL", "Stringized significator form: __KEY__ = \"STRING\""),
E("@%% KEY WS VALUE %% NL", "Multiline significator form"),
E("@%%! KEY WS STRING %% NL", "Multiline stringized significator form"),
E("@? NAME NL", "Set the current context name"),
E("@! INTEGER NL", "Set the current context line number"),
E("@(( ... )), @[[ ... ]],", "Extension markup; implementation provided by user"),
E("@{{ ... }}, @< ... >", ""),
]))

    usage.add(TableSection(
        'escapes',
        "Escape sequences",
        """\
Valid escape sequences are:
""", [
E("@\\0", "NUL, null", ord=0x0),
E("@\\a", "BEL, bell", ord=0x7),
E("@\\b", "BS, backspace", ord=0x8),
E("@\\B{BIN}", "freeform binary code BIN", ex="{1000001} for A"),
E("@\\dDDD", "three-digit decimal code DDD", ex="065 for A"),
E("@\\D{DEC}", "freeform decimal code DEC", ex="{65} for A"),
E("@\\e", "ESC, escape", ord=0x1b),
E("@\\f", "FF, form feed", ord=0xc),
E("@\\h", "DEL, delete", ord=0x7f),
E("@\\k", "ACK, acknowledge", ord=0x6),
E("@\\K", "NAK, negative acknowledge", ord=0x15),
E("@\\n", "LF, linefeed; newline", ord=0xa),
E("@\\N{NAME}", "Unicode character named NAME", ex="LATIN CAPITAL LETTER A"),
E("@\\oOOO", "three-digit octal code OOO", ex="101 for A"),
E("@\\O{OCT}", "freeform octal code OCT", ex="{101} for A"),
E("@\\qQQQQ", "four-digit quaternary code QQQQ", ex="1001 for A"),
E("@\\Q{QUA}", "freeform quaternary code QUA", ex="{1001} for A"),
E("@\\r", "CR, carriage return", ord=0xd),
E("@\\s", "SP, space", ord=0x20),
E("@\\S", "NBSP, no-break space", ord=0xa0),
E("@\\t", "HT, horizontal tab", ord=0x9),
E("@\\uHHHH", "16-bit hexadecimal Unicode HHHH", ex="0041 for A"),
E("@\\UHHHHHHHH", "32-bit hexadecimal Unicode HHHHHHHH", ex="00000041 for A"),
E("@\\v", "VT, vertical tab", ord=0xb),
E("@\\V{VS}", "VS, variation selector (1 .. 256)", ex="16 for emoji display"),
E("@\\w", "VS15, variation selector 15; text display", ord=0xfe0e),
E("@\\W", "VS16, variation selector 16; emoji display", ord=0xfe0f),
E("@\\xHH", "8-bit hexadecimal code HH", ex="41 for A"),
E("@\\X{HEX}", "freeform hexadecimal code HEX", ex="{41} for A"),
E("@\\y", "SUB, substitution", ord=0x1a),
E("@\\Y", "RC, replacement character", ord=0xfffd),
E("@\\z", "EOT, end of transmission", ord=0x4),
E("@\\Z", "ZWNBSP/BOM, zero-width no-break space/byte order mark", ord=0xfeff),
E("@\\,", "THSP, thin space", ord=0x2009),
E("@\\^C", "Control character C", ex="[ for ESC"),
E("@\\^{NAME}", "Escape named NAME (control character)", ex="ESC"),
E("@\\(", "Literal (", ord=0x28),
E("@\\)", "Literal )", ord=0x29),
E("@\\[", "Literal [", ord=0x5b),
E("@\\]", "Literal ]", ord=0x5d),
E("@\\{", "Literal {", ord=0x7b),
E("@\\}", "Literal }", ord=0x7d),
E("@\\<", "Literal <", ord=0x3c),
E("@\\>", "Literal >", ord=0x3e),
E("@\\\\", "Literal \\", ord=0x5c),
E("@\\'", "Literal '", ord=0x27),
E("@\\\"", "Literal \"", ord=0x22),
E("@\\?", "Literal ?", ord=0x3f),
]))

    usage.add(TableSection(
        'environ',
        "Environment variables",
        """\
The following environment variables are recognized (with corresponding
command line arguments):
""", [
E(em.OPTIONS_ENV, "Specify additional options to be included"),
E(em.CONFIG_ENV, "Specify configuration file(s) to load", arg='-c PATHS'),
E(em.PREFIX_ENV, "Specify the default prefix", arg='-p PREFIX'),
E(em.PSEUDO_ENV, "Specify name of pseudomodule", arg='-m NAME'),
E(em.FLATTEN_ENV, "Flatten empy pseudomodule if defined", arg='-f'),
E(em.RAW_ERRORS_ENV, "Show raw errors if defined", arg='-r'),
E(em.INTERACTIVE_ENV, "Enter interactive mode if defined", arg='-i'),
E(em.DELETE_ON_ERROR_ENV, "Delete output file on error", arg='-d'),
E(em.NO_PROXY_ENV, "Do not install sys.stdout proxy if defined", arg='-n'),
E(em.BUFFERING_ENV, "Buffer size (-1, 0, 1, or n)", arg='-b VALUE'),
E(em.BINARY_ENV, "Open input file as binary (for Python 2.x Unicode)", arg='-u'),
E(em.ENCODING_ENV, "Unicode both input and output encodings"),
E(em.INPUT_ENCODING_ENV, "Unicode input encoding"),
E(em.OUTPUT_ENCODING_ENV, "Unicode output encoding"),
E(em.ERRORS_ENV, "Unicode both input and output error handlers"),
E(em.INPUT_ERRORS_ENV, "Unicode input error handler"),
E(em.OUTPUT_ERRORS_ENV, "Unicode output error handler"),
]))

    usage.add(TableSection(
        'pseudo',
        "Pseudomodule attributes and functions",
        """\
The %s pseudomodule contains the following attributes and methods:
""" % usage.config.pseudomoduleName, [
E("version", "String representing EmPy version"),
E("compat", "List of applied Python compatibility features"),
E("executable", "The EmPy executable"),
E("argv", "EmPy script name and command line arguments"),
E("config", "The configuration for this interpreter"),
E("ok", "Is the interpreter still active?"),
E("error", "The last error to occur, or None"),
E("__init__(**kwargs)", "The interpreter constructor"),
E("__enter__/__exit__(...)", "Context manager support"),
E("reset()", "Reset the interpreter state"),
E("ready()", "Declare the interpreter ready"),
E("shutdown()", "Shutdown the interpreter"),
E("write(data)", "Write data to stream"),
E("writelines(lines)", "Write a sequence of lines to stream"),
E("flush()", "Flush the stream"),
E("serialize(object)", "Write a string version of the object"),
E("top() -> Stream", "Get the top-level stream"),
E("push()", "Push this interpreter on the stream stack"),
E("pop()", "Pop this interpreter off the stream stack"),
E("clear()", "Clear the interpreter stacks"),
E("go(fname, mode, [pre, [post]])", "Process a file by filename"),
E("interact()", "Enter interactive mode"),
E("file(file, [loc])", "Process a file object"),
E("fileLine(file, [loc])", "Process a file object in lines"),
E("fileChunk(file, buf, [loc]])", "Process a file in chunks"),
E("fileFull(file, [loc]])", "Process a file as a single chunk"),
E("process(command)", "Process a command"),
E("processAll(commands)", "Process a sequence of commands"),
E("identify()", "Identify top context as name, line"),
E("getContext()", "Return the current context"),
E("newContext(...)", "Return a new context with name and counts"),
E("pushContext(context)", "Push a context"),
E("popContext()", "Pop the current context"),
E("setContext(context)", "Replace the current context"),
E("setContextName(name)", "Set the name of the current context"),
E("setContextLine(line)", "Set the line number of the current context"),
E("setContextColumn(column)", "Set the column number of the current context"),
E("setContextData(...)", "Set any of the name, line, column number"),
E("restoreContext(context)", "Replace the top context with an existing one"),
E("clearFinalizers()", "Clear all finalizers"),
E("appendFinalizer(finalizer)", "Append function to be called at shutdown"),
E("prependFinalizer(finalizer)", "Prepend function to be called at shutdown"),
E("setFinalizers(finalizers)", "Set functions to be called at shutdown"),
E("getGlobals()", "Retrieve this interpreter's globals"),
E("setGlobals(dict)", "Set this interpreter's globals"),
E("updateGlobals(dict)", "Merge dictionary into interpreter's globals"),
E("clearGlobals()", "Start globals over anew"),
E("saveGlobals([deep])", "Save a copy of the globals"),
E("restoreGlobals([pop])", "Restore the most recently saved globals"),
E("include(file, [loc])", "Include filename or file-like object"),
E("expand(string, [loc])", "Explicitly expand string and return"),
E("defined(name, [loc])", "Find if the name is defined"),
E("lookup(name, [loc])", "Lookup the variable name"),
E("evaluate(expression, [loc])", "Evaluate the expression (and write?)"),
E("execute(statements, [loc])", "Execute the statements"),
E("single(source, [loc])", "Execute the expression or statement"),
E("atomic(name, value, [loc])", "Perform an atomic assignment"),
E("assign(name, value, [loc])", "Perform an arbitrary assignment"),
E("significate(key, [value])", "Significate the given key, value pair"),
E("quote(string)", "Quote prefixes in provided string and return"),
E("escape(data)", "Escape EmPy markup in data and return"),
E("flatten([keys])", "Flatten module contents to globals namespace"),
E("getPrefix()", "Get current prefix"),
E("setPrefix(char)", "Set new prefix"),
E("stopDiverting()", "Stop diverting; data sent directly to output"),
E("createDiversion(name)", "Create a diversion but do not divert to it"),
E("retrieveDiversion(name, [def])", "Retrieve the actual named diversion object"),
E("startDiversion(name)", "Start diverting to given diversion"),
E("playDiversion(name)", "Recall diversion and then eliminate it"),
E("replayDiversion(name)", "Recall diversion but retain it"),
E("dropDiversion(name)", "Drop diversion"),
E("playAllDiversions()", "Stop diverting and play all diversions"),
E("replayAllDiversions()", "Stop diverting and replay all diversions"),
E("dropAllDiversions()", "Stop diverting and drop all diversions"),
E("getCurrentDiversionName()", "Get the name of the current diversion"),
E("getAllDiversionNames()", "Get a sorted sequence of diversion names"),
E("isExistingDiversionName(name)", "Is this the name of a diversion?"),
E("getFilter()", "Get the first filter in the current chain"),
E("getLastFilter()", "Get the last filter in the current chain"),
E("getFilterCount()", "Get the number of filters in current chain"),
E("resetFilter()", "Reset filter; no filtering"),
E("setFilter(filter...)", "Install new filter(s), replacing any chain"),
E("prependFilter(filter)", "Prepend filter to beginning of current chain"),
E("appendFilter(filter)", "Append a filter to end of current chain"),
E("setFilterChain(filters)", "Install a new filter chain"),
E("areHooksEnabled()", "Return whether or not hooks are enabled"),
E("enableHooks()", "Enable hooks (default)"),
E("disableHooks()", "Disable hook invocation"),
E("getHooks()", "Get all the hooks"),
E("appendHook(hook)", "Append the given hook"),
E("prependHook(hook)", "Prepend the given hook"),
E("removeHook(hook)", "Remove an already-registered hook"),
E("clearHooks()", "Clear all hooks"),
E("invokeHook(_name, ...)", "Manually invoke hook"),
E("hasCallback()", "Is there a custom callback?"),
E("getCallback()", "Get custom callback"),
E("registerCallback(callback)", "Register custom callback"),
E("deregisterCallback()", "Deregister custom callback"),
E("invokeCallback(contents)", "Invoke the custom callback directly"),
E("defaultHandler(t, e, tb)", "The default error handler"),
E("getHandler()", "Get the current error handler (or None)"),
E("setHandler(handler, [eoe])", "Set the error handler"),
E("resetHandler([eoe])", "Reset the error handler"),
E("invokeHandler(t, e, tb)", "Manually invoke the error handler"),
E("initializeEmojiModules(names)", "Initialize the emoji modules"),
E("getEmojiModule(name)", "Get an abstracted emoji module"),
E("getEmojiModuleNames()", "Return the list of emoji module names"),
E("substituteEmoji(text)", "Do an emoji substitution"),
]))

    usage.add(TableSection(
        'constructor',
        "Keyword arguments for the Interpreter constructor",
        """\
The following keyword arguments for the Interpreter constructor are supported
(defaults in brackets).  All arguments have meaningful defaults:
""", [
E("argv", "The system arguments to use ['<->']"),
E("callback", "A custom callback to register [None]"),
E("config", "The configuration object [default]"),
E("core", "The core to use for this interpreter [Core()]"),
E("dispatcher", "Dispatch errors or raise to caller? [True]"),
E("executable", "The path to the EmPy executable [\".../em.py\"]"),
E("extension", "The extension to automatically install [None]"),
E("filespec", "A 3-tuple of the input filename, mode, and buffering [None]"),
E("filters", "The list of filters to install [[]]"),
E("finalizers", "The list of finalizers to installed [[]]"),
E("globals", "The globals dictionary to use [{}]"),
E("handler", "The error handler to use [default]"),
E("hooks", "The list of hooks to install [[]]"),
E("ident", "The identifier for the interpreter [None]"),
E("immediately", "Declare the interpreter ready after initialization? [True]"),
E("input", "The input file to use for interactivity [sys.stdin]"),
E("output", "The output file to use [sys.stdout]"),
E("root", "The root interpreter context filename ['<root>']"),
]))

    usage.add(ConfigSection(
        'variables',
        "Configuration variable attributes",
        """\
The following configuration variable attributes are supported (defaults in
brackets, with dictionaries being summarized with their length):
""",
        usage.config))

    usage.add(TableSection(
        'methods',
        "Configuration methods",
        """\
The configuration instance contains the following methods:
""", [
E("initialize()", "Initialize the instance"),
E("shutdown()", "Shutdown the instance"),
E("isInitialize()", "Is this configuration initialized?"),
E("pauseIfRequired()", "Pause if required"),
E("check(in, out)", "Check file settings"),
E("has(name)", "Is this variable defined?"),
E("get(name[, default])", "Get this variable value"),
E("set(name, value)", "Set this variable"),
E("update(**kwargs)", "Update with dictionary"),
E("clone([deep])", "Clone (optionally deep) this configuration"),
E("run(statements)", "Run configuration commands"),
E("load(filename, [required])", "Load configuration file"),
E("path(filename, [required])", "Load configuration file(s) path"),
E("hasEnvironment(name)", "Is this environment variable defined?"),
E("environment(name, ...)", "Get the enviroment variable value"),
E("hasDefaultPrefix()", "Is the prefix the default?"),
E("hasFullBuffering()", "Is the buffering set to full?"),
E("hasNoBuffering()", "Is the buffering set to none?"),
E("hasLineBuffering()", "Is the buffering set to line?"),
E("hasFixedBuffering()", "Is the buffering set to fixed?"),
E("createFactory([tokens])", "Create token factory"),
E("adjustFactory()", "Adjust token factory for non-default prefix"),
E("getFactory([tokens], [force])", "Get a token factory"),
E("resetFactory()", "Clear the current token factory"),
E("hasBinary()", "Is binary (Unicode) support enabled?"),
E("enableBinary(...)", "Enable binary (Unicode) support"),
E("disableBinary()", "Disable binary (Unicode) support"),
E("isDefaultEncodingErrors()", "Is encoding/errors the default?"),
E("getDefaultEncoding()", "Get the default file encoding"),
E("open(filename, [mode], ...)", "Open a file"),
E("significatorReString()", "Regular expression string for significators"),
E("significatorRe()", "Regular expression pattern for significators"),
E("significatorFor(key)", "Significator variable name for key"),
E("setContextFormat(format)", "Set the context format"),
E("renderContext(context)", "Render context using format"),
E("calculateIconsSignature()", "Calculate icons signature"),
E("signIcons()", "Sign the icons dictionary"),
E("transmogrifyIcons()", "Process the icons dictionary"),
E("validateIcons()", "Process the icons dictionaray if necessary"),
E("intializeEmojiModules([names])", "Initialize emoji module support"),
E("substituteEmoji(text)", "Substitute emoji"),
E("isSuccessCode(code)", "Does this exit code indicate success?"),
E("isExitError(error)", "Is this exception instance an exit?"),
E("errorToExitCode(error)", "Return exit code for exception instance"),
E("isNotAnError(error)", "Is this exception instance not an error?"),
E("formatError(error[, p, s])", "Render an error string from instance"),
]))

    usage.add(TableSection(
        'hooks',
        "Hook methods",
        """\
The following hook methods are supported.  The return values are ignored except
for the `pre...` methods which, when they return a true value, signal that the
following token handling should be skipped:
""", [
E("atInstallProxy(proxy, new)", "Proxy being installed"),
E("atUninstallProxy(proxy, new)", "Proxy being uninstalled"),
E("atStartup()", "Interpreter started up"),
E("atReady()", "Interpreter ready"),
E("atFinalize()", "Interpreter finalizing"),
E("atShutdown()", "Interpreter shutting down"),
E("atParse(scanner, loc)", "Interpreter parsing"),
E("atToken(token)", "Interpreter expanding token"),
E("atHandle(info, fatal, contexts)", "Interpreter encountered error"),
E("atInteract()", "Interpreter going interactive"),
E("pushContext(context)", "Context being pushed"),
E("popContext(context)", "Context was popped"),
E("setContext(context)", "Context was set or modified"),
E("restoreContext(context)", "Context was restored"),
E("prePrefix()", "Pre prefix token"),
E("preWhitespace()", "Pre whitespace token"),
E("preString(string)", "Pre string literal token"),
E("preLineComment(comment)", "Pre line comment"),
E("postLineComment()", "Post line comment"),
E("preInlineComment(comment)", "Pre inline comment"),
E("postInlineComment()", "Post inline comment"),
E("preBackquote(literal)", "Pre backquote literal"),
E("postBackquote()", "Post backquote literal"),
E("preSignificator(key, value, s)", "Pre significator"),
E("postSignificator()", "post significator"),
E("preContextName(name)", "Pre context name"),
E("postContextName()", "Post context name"),
E("preContextLine(line)", "Pre context line"),
E("postContextLine()", "Post context line"),
E("preExpression(pairs, except, loc)", "Pre expression"),
E("postExpression(result)", "Post expression"),
E("preSimple(code, sub, loc)", "Pre simple expression"),
E("postSimple(result)", "Post simple expression"),
E("preInPlace(code, loc)", "Pre in-place expression"),
E("postInPlace(result)", "Post in-place expression"),
E("preStatement(code, loc)", "Pre statement"),
E("postStatement()", "Post statement"),
E("preControl(type, rest, loc)", "Pre control"),
E("postControl()", "Post control"),
E("preEscape(code)", "Pre escape"),
E("postEscape()", "Post escape"),
E("preDiacritic(code)", "Pre diacritic"),
E("postDiacritic()", "Post diacritic"),
E("preIcon(code)", "Pre icon"),
E("postIcon()", "Post icon"),
E("preEmoji(name)", "Pre emoji"),
E("postEmoji()", "Post emoji"),
E("preCustom(contents)", "Pre custom"),
E("postCustom()", "Post custom"),
E("beforeProcess(command, n)", "Before command processing"),
E("afterProcess()", "After command processing"),
E("beforeInclude(file, loc, name)", "Before file inclusion"),
E("afterInclude()", "After file inclusion"),
E("beforeExpand(string, loc, name)", "Before expand call"),
E("afterExpand(result)", "After expand call"),
E("beforeTokens(tokens, loc)", "Before processing tokens"),
E("afterTokens(result)", "After processing tokens"),
E("beforeFileLines(file, loc)", "Before reading file by lines"),
E("afterFileLines()", "After reading file by lines"),
E("beforeFileChunks(file, loc)", "Before reading file by chunks"),
E("afterFileChunks()", "After reading file by chunks"),
E("beforeFileFull(file, loc)", "Before reading file in full"),
E("afterFilFull()", "After reading file in full"),
E("beforeString(string, loc)", "Before processing string"),
E("afterString()", "After processing string"),
E("beforeQuote(string)", "Before quoting string"),
E("afterQuote()", "After quoting string"),
E("beforeEscape(string)", "Before escaping string"),
E("afterEscape()", "After escaping string"),
E("beforeSignificate(key, value, loc)", "Before significator"),
E("afterSignificate()", "After significator"),
E("beforeCallback(contents)", "Before custom callback"),
E("afterCallback()", "Before custom callback"),
E("beforeAtomic(name, value, loc)", "Before atomic assignment"),
E("afterAtomic()", "After atomic assignment"),
E("beforeMulti(names, values, loc)", "Before complex assignment"),
E("afterMulti()", "After complex assignment"),
E("beforeImport(name, loc)", "Before module import"),
E("afterImport()", "After module import"),
E("beforeFunctional(code, lists, loc)", "Before functional expression"),
E("afterFunctional(result)", "After functional expression"),
E("beforeEvaluate(code, loc, write)", "Before expression evaluation"),
E("afterEvaluate(result)", "After expression evaluation"),
E("beforeExecute(statements, loc)", "Before statement execution"),
E("afterExecute()", "After statement execution"),
E("beforeSingle(source, loc)", "Before single execution"),
E("afterSingle(result)", "After single execution"),
E("beforeFinalizer(final)", "Before finalizer processing"),
E("afterFinalizer()", "After finalizer processing"),
]))

    usage.add(MappingSection(
        'named',
        "Named escapes",
        "The following named escapes (control codes and control-like characters)\n(`@\\^{...}`) are supported:\n",
        usage.config.controls,
        ControlEntry))

    usage.add(MappingSection(
        'diacritics',
        "Diacritic combiners",
        "The following diacritic combining characters (`@^C...`) are supported:\n",
        usage.config.diacritics))

    usage.add(MappingSection(
        'icons',
        "Icons",
        "The following icon sequences (`@|...`) are supported:\n",
        usage.config.icons,
        NameEntry))

    usage.add(MappingSection(
        'emojis',
        "Custom emojis",
        "The following custom emojis have been made available:\n",
        usage.config.emojis,
        NameEntry))

    usage.add(HelpSection(
        'topics',
        "This list of topics",
        """\
Need more help?  Add more -h options (-hh, -hhh) for more help.  Use -H <topic>
for help on a specific topic, or specify a comma-separated list of topics.  Try
`default` (-h) for default help, `more` (-hh) for more common topics, `all`
(-hhh) for all help topics, or `topics` for this list.  Use -V for version
information, -W for version and system information, or -Z for all debug
details.  Available topics:
""",
        usage))

    return payload

#
# main
#

def main():
    topics = sys.argv[1:]
    if not topics:
        topics = ['default']
    usage = Usage()
    usage.show(topics)

if __name__ == '__main__': main()
